"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const BufferStream_1 = tslib_1.__importDefault(require("../../common/utils/BufferStream"));
const utils_1 = require("../../common/utils/utils");
const TableBatchOrchestrator_1 = tslib_1.__importDefault(require("../batch/TableBatchOrchestrator"));
const TableStorageContext_1 = tslib_1.__importDefault(require("../context/TableStorageContext"));
const NormalizedEntity_1 = require("../entity/NormalizedEntity");
const NotImplementedError_1 = tslib_1.__importDefault(require("../errors/NotImplementedError"));
const StorageErrorFactory_1 = tslib_1.__importDefault(require("../errors/StorageErrorFactory"));
const constants_1 = require("../utils/constants");
const utils_2 = require("../utils/utils");
const BaseHandler_1 = tslib_1.__importDefault(require("./BaseHandler"));
const TableBatchUtils_1 = tslib_1.__importDefault(require("../batch/TableBatchUtils"));
const to_readable_stream_1 = tslib_1.__importDefault(require("to-readable-stream"));
/**
 * TODO:
 * 1. Check Accept for every API
 * 2. Check Prefer for every API
 */
class TableHandler extends BaseHandler_1.default {
    async create(tableProperties, options, context) {
        const tableContext = new TableStorageContext_1.default(context);
        const accept = this.getAndCheckPayloadFormat(tableContext);
        const account = this.getAndCheckAccountName(tableContext);
        const table = tableProperties.tableName; // Table name is in request body instead of URL
        if (table === undefined) {
            throw StorageErrorFactory_1.default.getTableNameEmpty(context);
        }
        // validate table name
        if (table !== undefined) {
            utils_2.validateTableName(context, table);
        }
        const tableModel = {
            account,
            table
        };
        await this.metadataStore.createTable(context, tableModel);
        const response = {
            clientRequestId: options.requestId,
            requestId: tableContext.contextID,
            version: constants_1.TABLE_API_VERSION,
            date: context.startTime,
            statusCode: 201
        };
        response.tableName = table;
        utils_2.updateTableOptionalOdataAnnotationsForResponse(response, account, table, this.getOdataAnnotationUrlPrefix(tableContext, account), accept);
        this.updateResponsePrefer(response, tableContext);
        this.updateResponseAccept(tableContext, accept);
        return response;
    }
    async delete(_table, options, context) {
        const tableContext = new TableStorageContext_1.default(context);
        const account = this.getAndCheckAccountName(tableContext);
        const table = this.getAndCheckTableName(tableContext);
        const accept = this.getAndCheckPayloadFormat(tableContext);
        await this.metadataStore.deleteTable(context, table, account);
        const response = {
            clientRequestId: options.requestId,
            requestId: tableContext.contextID,
            version: constants_1.TABLE_API_VERSION,
            date: context.startTime,
            statusCode: 204
        };
        this.updateResponseAccept(tableContext, accept);
        return response;
    }
    async query(options, context) {
        var _a;
        const tableContext = new TableStorageContext_1.default(context);
        const account = this.getAndCheckAccountName(tableContext);
        const accept = this.getAndCheckPayloadFormat(tableContext);
        const [tableResult, nextTableName] = await this.metadataStore.queryTable(context, account, (_a = options.queryOptions) === null || _a === void 0 ? void 0 : _a.top, options.nextTableName);
        const response = {
            clientRequestId: options.requestId,
            requestId: tableContext.contextID,
            version: constants_1.TABLE_API_VERSION,
            date: context.startTime,
            statusCode: 200,
            xMsContinuationNextTableName: nextTableName,
            value: []
        };
        const prefix = this.getOdataAnnotationUrlPrefix(tableContext, account);
        const annotation = utils_2.getTableOdataAnnotationsForResponse(account, "", prefix);
        if (accept === constants_1.MINIMAL_METADATA_ACCEPT || accept === constants_1.FULL_METADATA_ACCEPT) {
            response.odatametadata = annotation.odatametadata;
        }
        response.value = tableResult.map((item) => utils_2.getTablePropertiesOdataAnnotationsForResponse(item.table, account, prefix, accept));
        this.updateResponseAccept(tableContext, accept);
        return response;
    }
    // TODO: Filter odata types per accept settings
    async insertEntity(_tableName, options, context) {
        const tableContext = new TableStorageContext_1.default(context);
        const account = this.getAndCheckAccountName(tableContext);
        const table = this.getAndCheckTableName(tableContext);
        const accept = this.getAndCheckPayloadFormat(tableContext);
        const prefer = this.getAndCheckPreferHeader(tableContext);
        // curently unable to use checking functions as the partitionKey
        // and rowKey are not coming through the context.
        // const partitionKey = this.getAndCheckPartitionKey(tableContext);
        // const rowKey = this.getAndCheckRowKey(tableContext);
        if (!options.tableEntityProperties ||
            // rowKey and partitionKey may be empty string
            options.tableEntityProperties.PartitionKey === null ||
            options.tableEntityProperties.RowKey === null) {
            throw StorageErrorFactory_1.default.getPropertiesNeedValue(context);
        }
        const entity = {
            PartitionKey: options.tableEntityProperties.PartitionKey,
            RowKey: options.tableEntityProperties.RowKey,
            properties: options.tableEntityProperties,
            lastModifiedTime: context.startTime,
            eTag: utils_1.newTableEntityEtag(context.startTime)
        };
        let nomarlizedEntity;
        try {
            nomarlizedEntity = new NormalizedEntity_1.NormalizedEntity(entity);
            nomarlizedEntity.normalize();
        }
        catch (e) {
            this.logger.error(`TableHandler:insertEntity() ${e.name} ${JSON.stringify(e.stack)}`, context.contextID);
            throw StorageErrorFactory_1.default.getInvalidInput(context);
        }
        await this.metadataStore.insertTableEntity(context, table, account, entity);
        const response = {
            clientRequestId: options.requestId,
            requestId: tableContext.contextID,
            version: constants_1.TABLE_API_VERSION,
            date: context.startTime,
            statusCode: 201,
            eTag: entity.eTag
        };
        if (prefer === constants_1.RETURN_CONTENT || prefer === undefined) {
            const body = {};
            const annotation = utils_2.getEntityOdataAnnotationsForResponse(account, table, this.getOdataAnnotationUrlPrefix(tableContext, account), options.tableEntityProperties.PartitionKey, options.tableEntityProperties.RowKey, accept);
            if (accept === constants_1.MINIMAL_METADATA_ACCEPT) {
                body["odata.metadata"] = annotation.odatametadata;
                body["odata.etag"] = entity.eTag;
            }
            if (accept === constants_1.FULL_METADATA_ACCEPT) {
                body["odata.metadata"] = annotation.odatametadata;
                body["odata.type"] = annotation.odatatype;
                body["odata.id"] = annotation.odataid;
                body["odata.etag"] = entity.eTag;
                body["odata.editLink"] = annotation.odataeditLink;
            }
            // for (const key of Object.keys(entity.properties)) {
            //   body[key] = entity.properties[key];
            // }
            // response.body = new BufferStream(Buffer.from(JSON.stringify(body)));
            const rawResponse = nomarlizedEntity.toResponseString(accept, body);
            this.logger.debug(`TableHandler:insertEntity() Raw response string is ${JSON.stringify(rawResponse)}`, context.contextID);
            response.body = new BufferStream_1.default(Buffer.from(rawResponse));
        }
        this.updateResponseAccept(tableContext, accept);
        this.updateResponsePrefer(response, tableContext);
        return response;
    }
    // TODO: Create data structures to hold entity properties and support serialize, merge, deserialize, filter
    // Note: Batch is using the partition key and row key args, handler receives these values from middleware via
    // context
    async updateEntity(_table, partitionKey, rowKey, options, context) {
        const tableContext = new TableStorageContext_1.default(context);
        const account = this.getAndCheckAccountName(tableContext);
        const table = this.getAndCheckTableName(tableContext);
        partitionKey = partitionKey || this.getAndCheckPartitionKey(tableContext);
        rowKey = rowKey || this.getAndCheckRowKey(tableContext);
        const ifMatch = options.ifMatch;
        if (!options.tableEntityProperties) {
            throw StorageErrorFactory_1.default.getPropertiesNeedValue(context);
        }
        if (options.tableEntityProperties.PartitionKey !== partitionKey ||
            options.tableEntityProperties.RowKey !== rowKey) {
            this.logger.warn(`TableHandler:updateEntity() Incoming PartitionKey:${partitionKey} RowKey:${rowKey} in URL parameters don't align with entity body PartitionKey:${options.tableEntityProperties.PartitionKey} RowKey:${options.tableEntityProperties.RowKey}.`);
        }
        // Test if etag is available
        // this is considered an upsert if no etag header, an empty header is an error.
        // https://docs.microsoft.com/en-us/rest/api/storageservices/insert-or-replace-entity
        if (ifMatch === "") {
            throw StorageErrorFactory_1.default.getPreconditionFailed(context);
        }
        if ((options === null || options === void 0 ? void 0 : options.ifMatch) && options.ifMatch !== "*") {
            if (utils_1.checkEtagIsInvalidFormat(options.ifMatch)) {
                throw StorageErrorFactory_1.default.getInvalidOperation(context);
            }
        }
        const eTag = utils_1.newTableEntityEtag(context.startTime);
        // Entity, which is used to update an existing entity
        const entity = {
            PartitionKey: partitionKey,
            RowKey: rowKey,
            properties: options.tableEntityProperties,
            lastModifiedTime: context.startTime,
            eTag
        };
        let nomarlizedEntity;
        try {
            nomarlizedEntity = new NormalizedEntity_1.NormalizedEntity(entity);
            nomarlizedEntity.normalize();
        }
        catch (e) {
            this.logger.error(`TableHandler:updateEntity() ${e.name} ${JSON.stringify(e.stack)}`, context.contextID);
            throw StorageErrorFactory_1.default.getInvalidInput(context);
        }
        await this.metadataStore.insertOrUpdateTableEntity(context, table, account, entity, ifMatch);
        // Response definition
        const response = {
            clientRequestId: options.requestId,
            requestId: tableContext.contextID,
            version: constants_1.TABLE_API_VERSION,
            date: context.startTime,
            eTag,
            statusCode: 204
        };
        return response;
    }
    async mergeEntity(_table, partitionKey, rowKey, options, context) {
        const tableContext = new TableStorageContext_1.default(context);
        const account = this.getAndCheckAccountName(tableContext);
        const table = this.getAndCheckTableName(tableContext);
        partitionKey = partitionKey || this.getAndCheckPartitionKey(tableContext);
        rowKey = rowKey || this.getAndCheckRowKey(tableContext);
        if (!options.tableEntityProperties) {
            throw StorageErrorFactory_1.default.getPropertiesNeedValue(context);
        }
        if ((options === null || options === void 0 ? void 0 : options.ifMatch) && options.ifMatch !== "*" && options.ifMatch !== "") {
            if (utils_1.checkEtagIsInvalidFormat(options.ifMatch)) {
                throw StorageErrorFactory_1.default.getInvalidOperation(context);
            }
        }
        if (options.tableEntityProperties.PartitionKey !== partitionKey ||
            options.tableEntityProperties.RowKey !== rowKey) {
            this.logger.warn(`TableHandler:mergeEntity() Incoming PartitionKey:${partitionKey} RowKey:${rowKey} in URL parameters don't align with entity body PartitionKey:${options.tableEntityProperties.PartitionKey} RowKey:${options.tableEntityProperties.RowKey}.`);
        }
        const eTag = utils_1.newTableEntityEtag(context.startTime);
        const entity = {
            PartitionKey: partitionKey,
            RowKey: rowKey,
            properties: options.tableEntityProperties,
            lastModifiedTime: context.startTime,
            eTag
        };
        let nomarlizedEntity;
        try {
            nomarlizedEntity = new NormalizedEntity_1.NormalizedEntity(entity);
            nomarlizedEntity.normalize();
        }
        catch (e) {
            this.logger.error(`TableHandler:mergeEntity() ${e.name} ${JSON.stringify(e.stack)}`, context.contextID);
            throw StorageErrorFactory_1.default.getInvalidInput(context);
        }
        await this.metadataStore.insertOrMergeTableEntity(context, table, account, entity, options.ifMatch);
        const response = {
            clientRequestId: options.requestId,
            requestId: tableContext.contextID,
            version: constants_1.TABLE_API_VERSION,
            date: context.startTime,
            statusCode: 204,
            eTag
        };
        return response;
    }
    async deleteEntity(_table, partitionKey, rowKey, ifMatch, options, context) {
        const tableContext = new TableStorageContext_1.default(context);
        const accountName = tableContext.account;
        partitionKey = partitionKey || tableContext.partitionKey; // Get partitionKey from context
        rowKey = rowKey || tableContext.rowKey; // Get rowKey from context
        if (!partitionKey || !rowKey) {
            throw StorageErrorFactory_1.default.getPropertiesNeedValue(context);
        }
        if (ifMatch === "" || ifMatch === undefined) {
            throw StorageErrorFactory_1.default.getPreconditionFailed(context);
        }
        if (ifMatch !== "*" && utils_1.checkEtagIsInvalidFormat(ifMatch)) {
            throw StorageErrorFactory_1.default.getInvalidOperation(context);
        }
        // currently the props are not coming through as args, so we take them from the table context
        await this.metadataStore.deleteTableEntity(context, tableContext.tableName, accountName, partitionKey, rowKey, ifMatch);
        return {
            statusCode: 204,
            date: tableContext.startTime,
            clientRequestId: options.requestId,
            requestId: tableContext.contextID,
            version: constants_1.TABLE_API_VERSION
        };
    }
    async queryEntities(_table, options, context) {
        var _a, _b;
        const tableContext = new TableStorageContext_1.default(context);
        const table = this.getAndCheckTableName(tableContext);
        const account = this.getAndCheckAccountName(tableContext);
        const accept = this.getAndCheckPayloadFormat(tableContext);
        const [result, nextPartitionKey, nextRowKey] = await this.metadataStore.queryTableEntities(context, account, table, options.queryOptions || {}, options.nextPartitionKey, options.nextRowKey);
        const response = {
            clientRequestId: options.requestId,
            requestId: tableContext.contextID,
            version: constants_1.TABLE_API_VERSION,
            date: context.startTime,
            xMsContinuationNextPartitionKey: nextPartitionKey,
            xMsContinuationNextRowKey: nextRowKey,
            statusCode: 200
        };
        let selectSet;
        const selectArray = (_b = (_a = options.queryOptions) === null || _a === void 0 ? void 0 : _a.select) === null || _b === void 0 ? void 0 : _b.split(",").filter((item) => {
            return typeof item === "string" && item.length > 0;
        }).map((item) => item.trim());
        if (selectArray && selectArray.length > 0) {
            selectSet = new Set(selectArray);
        }
        const entities = [];
        const odataPrefix = this.getOdataAnnotationUrlPrefix(tableContext, account);
        result.forEach((element) => {
            const entity = {};
            const annotation = utils_2.getEntityOdataAnnotationsForResponse(account, table, odataPrefix, element.PartitionKey, element.RowKey, accept);
            if (accept === constants_1.MINIMAL_METADATA_ACCEPT ||
                accept === constants_1.FULL_METADATA_ACCEPT) {
                entity["odata.etag"] = element.eTag;
            }
            if (accept === constants_1.FULL_METADATA_ACCEPT) {
                entity["odata.type"] = annotation.odatatype;
                entity["odata.id"] = annotation.odataid;
                entity["odata.editLink"] = annotation.odataeditLink;
            }
            const nomarlizedEntity = new NormalizedEntity_1.NormalizedEntity(element);
            entities.push(nomarlizedEntity.toResponseString(accept, entity, selectSet));
        });
        const odatametadata = utils_2.getEntityOdataAnnotationsForResponse(account, table, odataPrefix, "", "", accept).odatametadata;
        const odatametadataPariString = odatametadata
            ? `"odata.metadata":${JSON.stringify(odatametadata)},`
            : "";
        const body = `{${odatametadataPariString}"value":[${entities.join(",")}]}`;
        response.body = new BufferStream_1.default(Buffer.from(body));
        this.logger.debug(`TableHandler:queryEntities() Raw response string is ${JSON.stringify(body)}`, context.contextID);
        this.updateResponseAccept(tableContext, accept);
        return response;
    }
    async queryEntitiesWithPartitionAndRowKey(_table, partitionKey, rowKey, options, context) {
        var _a, _b;
        const tableContext = new TableStorageContext_1.default(context);
        const account = this.getAndCheckAccountName(tableContext);
        const table = _table ? _table : this.getAndCheckTableName(tableContext);
        partitionKey = partitionKey || this.getAndCheckPartitionKey(tableContext);
        rowKey = rowKey || this.getAndCheckRowKey(tableContext);
        const accept = this.getAndCheckPayloadFormat(tableContext);
        const entity = await this.metadataStore.queryTableEntitiesWithPartitionAndRowKey(context, table, account, partitionKey, rowKey);
        if (entity === undefined || entity === null) {
            throw StorageErrorFactory_1.default.getEntityNotFound(context);
        }
        const response = {
            statusCode: 200,
            date: tableContext.startTime,
            clientRequestId: options.requestId,
            requestId: context.contextID,
            version: constants_1.TABLE_API_VERSION
        };
        const body = {};
        const annotation = utils_2.getEntityOdataAnnotationsForResponse(account, table, this.getOdataAnnotationUrlPrefix(tableContext, account), partitionKey, rowKey, accept);
        if (accept === constants_1.MINIMAL_METADATA_ACCEPT) {
            body["odata.metadata"] = annotation.odatametadata;
            body["odata.etag"] = entity.eTag;
        }
        if (accept === constants_1.FULL_METADATA_ACCEPT) {
            body["odata.metadata"] = annotation.odatametadata;
            body["odata.type"] = annotation.odatatype;
            body["odata.id"] = annotation.odataid;
            body["odata.etag"] = entity.eTag;
            body["odata.editLink"] = annotation.odataeditLink;
        }
        let selectSet;
        const selectArray = (_b = (_a = options.queryOptions) === null || _a === void 0 ? void 0 : _a.select) === null || _b === void 0 ? void 0 : _b.split(",").filter((item) => {
            return typeof item === "string" && item.length > 0;
        }).map((item) => item.trim());
        if (selectArray && selectArray.length > 0) {
            selectSet = new Set(selectArray);
        }
        const nomarlizedEntity = new NormalizedEntity_1.NormalizedEntity(entity);
        const rawResponse = nomarlizedEntity.toResponseString(accept, body, selectSet);
        response.body = new BufferStream_1.default(Buffer.from(rawResponse));
        this.logger.debug(`TableHandler:queryEntities() Raw response string is ${JSON.stringify(rawResponse)}`, context.contextID);
        this.updateResponseAccept(tableContext, accept);
        return response;
    }
    async mergeEntityWithMerge(table, partitionKey, rowKey, options, context) {
        return this.mergeEntity(table, partitionKey, rowKey, options, context);
    }
    async getAccessPolicy(table, options, context) {
        // e.g
        // const tableContext = new TableStorageContext(context);
        // const accountName = tableContext.account;
        // const tableName = tableContext.tableName; // Get tableName from context
        // TODO
        throw new NotImplementedError_1.default(context);
    }
    async setAccessPolicy(table, options, context) {
        // e.g
        // const tableContext = new TableStorageContext(context);
        // const accountName = tableContext.account;
        // const tableName = tableContext.tableName; // Get tableName from context
        // TODO
        throw new NotImplementedError_1.default(context);
    }
    /**
     * Processes an entity group transaction request / batch request
     *
     * @param {NodeJS.ReadableStream} body
     * @param {string} multipartContentType
     * @param {number} contentLength
     * @param {Models.TableBatchOptionalParams} options
     * @param {Context} context
     * @return {*}  {Promise<Models.TableBatchResponse>}
     * @memberof TableHandler
     */
    async batch(body, multipartContentType, contentLength, options, context) {
        var _a, _b;
        const tableCtx = new TableStorageContext_1.default(context);
        const contentTypeResponse = (_b = (_a = tableCtx.request) === null || _a === void 0 ? void 0 : _a.getHeader("content-type")) === null || _b === void 0 ? void 0 : _b.replace("batch", "batchresponse");
        const tableBatchManager = new TableBatchOrchestrator_1.default(tableCtx, this);
        const requestBody = await TableBatchUtils_1.default.StreamToString(body);
        this.logger.debug(`TableHandler:batch() Raw request string is ${JSON.stringify(requestBody)}`, context.contextID);
        const response = await tableBatchManager.processBatchRequestAndSerializeResponse(requestBody);
        this.logger.debug(`TableHandler:batch() Raw response string is ${JSON.stringify(response)}`, context.contextID);
        // need to convert response to NodeJS.ReadableStream
        body = to_readable_stream_1.default(response);
        return {
            contentType: contentTypeResponse,
            requestId: tableCtx.contextID,
            version: constants_1.TABLE_API_VERSION,
            date: context.startTime,
            statusCode: 202,
            body
        };
    }
    getOdataAnnotationUrlPrefix(tableContext, account) {
        // TODO: Get protocol, host and port from Azurite server instance
        let protocol = "http";
        let host = `${constants_1.DEFAULT_TABLE_SERVER_HOST_NAME}:${constants_1.DEFAULT_TABLE_LISTENING_PORT}/${account}`;
        if (tableContext.request !== undefined) {
            host = `${tableContext.request.getHeader("host")}/${account}` || host;
            protocol = tableContext.request.getProtocol();
        }
        return `${protocol}://${host}`;
    }
    getAndCheckPayloadFormat(context) {
        const format = utils_2.getPayloadFormat(context);
        if (format !== constants_1.NO_METADATA_ACCEPT &&
            format !== constants_1.MINIMAL_METADATA_ACCEPT &&
            format !== constants_1.FULL_METADATA_ACCEPT) {
            throw StorageErrorFactory_1.default.getAtomFormatNotSupported(context);
        }
        return format;
    }
    getAndCheckPreferHeader(context) {
        const prefer = context.request.getHeader(constants_1.HeaderConstants.PREFER);
        return prefer;
    }
    getAndCheckAccountName(context) {
        const account = context.account;
        if (account === undefined) {
            throw StorageErrorFactory_1.default.getAccountNameEmpty(context);
        }
        return account;
    }
    getAndCheckTableName(context) {
        const table = context.tableName;
        if (table === undefined) {
            throw StorageErrorFactory_1.default.getTableNameEmpty(context);
        }
        return table;
    }
    getAndCheckPartitionKey(context) {
        const partitionKey = context.partitionKey;
        if (partitionKey === undefined) {
            throw StorageErrorFactory_1.default.getTableNameEmpty(context);
        }
        return partitionKey;
    }
    getAndCheckRowKey(context) {
        const rowKey = context.rowKey;
        if (rowKey === undefined) {
            throw StorageErrorFactory_1.default.getTableNameEmpty(context);
        }
        return rowKey;
    }
    updateResponseAccept(context, accept) {
        if (accept !== undefined) {
            context.response.setContentType(accept);
        }
        return context;
    }
    updateResponsePrefer(response, context) {
        const prefer = context.request.getHeader(constants_1.HeaderConstants.PREFER);
        if (prefer === constants_1.RETURN_NO_CONTENT) {
            response.statusCode = 204;
            response.preferenceApplied = constants_1.RETURN_NO_CONTENT;
        }
        if (prefer === constants_1.RETURN_CONTENT || prefer === undefined) {
            response.statusCode = 201;
            response.preferenceApplied = constants_1.RETURN_CONTENT;
        }
        return response;
    }
}
exports.default = TableHandler;
//# sourceMappingURL=TableHandler.js.map