"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const fs_1 = require("fs");
const lokijs_1 = tslib_1.__importDefault(require("lokijs"));
const NotImplementedError_1 = tslib_1.__importDefault(require("../errors/NotImplementedError"));
const StorageErrorFactory_1 = tslib_1.__importDefault(require("../errors/StorageErrorFactory"));
const constants_1 = require("../utils/constants");
const utils_1 = require("../utils/utils");
class LokiTableMetadataStore {
    constructor(lokiDBPath) {
        this.lokiDBPath = lokiDBPath;
        this.TABLES_COLLECTION = "$TABLES_COLLECTION$";
        this.initialized = false;
        this.closed = false;
        this.db = new lokijs_1.default(lokiDBPath, {
            autosave: true,
            autosaveInterval: 5000
        });
    }
    async init() {
        await new Promise((resolve, reject) => {
            fs_1.stat(this.lokiDBPath, (statError, stats) => {
                if (!statError) {
                    this.db.loadDatabase({}, (dbError) => {
                        if (dbError) {
                            reject(dbError);
                        }
                        else {
                            resolve();
                        }
                    });
                }
                else {
                    // when DB file doesn't exist, ignore the error because following will re-create the file
                    resolve();
                }
            });
        });
        // Create tables collection if not exists
        if (this.db.getCollection(this.TABLES_COLLECTION) === null) {
            this.db.addCollection(this.TABLES_COLLECTION, {
                // Optimization for indexing and searching
                // https://rawgit.com/techfort/LokiJS/master/jsdoc/tutorial-Indexing%20and%20Query%20performance.html
                indices: ["account", "table"]
            }); // Optimize for find operation
        }
        await new Promise((resolve, reject) => {
            this.db.saveDatabase((err) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve();
                }
            });
        });
        this.initialized = true;
        this.closed = false;
    }
    async close() {
        await new Promise((resolve, reject) => {
            this.db.close((err) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve();
                }
            });
        });
        this.closed = true;
    }
    isInitialized() {
        return this.initialized;
    }
    isClosed() {
        return this.closed;
    }
    async createTable(context, tableModel) {
        // Check for table entry in the table registry collection
        const coll = this.db.getCollection(this.TABLES_COLLECTION);
        const doc = coll.findOne({
            account: tableModel.account,
            table: tableModel.table
        });
        // If the metadata exists, we will throw getTableAlreadyExists error
        if (doc) {
            throw StorageErrorFactory_1.default.getTableAlreadyExists(context);
        }
        coll.insert(tableModel);
        // now we create the collection to represent the table using a unique string
        const tableCollectionName = this.getTableCollectionName(tableModel.account, tableModel.table);
        const extentColl = this.db.getCollection(tableCollectionName);
        if (extentColl) {
            this.db.removeCollection(tableCollectionName);
        }
        this.db.addCollection(tableCollectionName, {
            // Optimization for indexing and searching
            // https://rawgit.com/techfort/LokiJS/master/jsdoc/tutorial-Indexing%20and%20Query%20performance.html
            indices: ["PartitionKey", "RowKey"]
        }); // Optimize for find operation
    }
    async deleteTable(context, table, account) {
        // remove table reference from collection registry
        const coll = this.db.getCollection(this.TABLES_COLLECTION);
        const doc = coll.findOne({
            account,
            table
        });
        if (doc) {
            coll.remove(doc);
        }
        else {
            throw StorageErrorFactory_1.default.ResourceNotFound(context);
        }
        const tableCollectionName = this.getTableCollectionName(account, table);
        const tableEntityCollection = this.db.getCollection(tableCollectionName);
        if (tableEntityCollection) {
            this.db.removeCollection(tableCollectionName);
        }
    }
    async queryTable(context, account, top = 1000, nextTable) {
        const coll = this.db.getCollection(this.TABLES_COLLECTION);
        const filter = { account };
        if (nextTable) {
            filter.table = { $gte: nextTable };
        }
        const docList = coll
            .chain()
            .find(filter)
            .simplesort("table")
            .limit(top + 1)
            .data();
        let nextTableName;
        if (docList.length > top) {
            const tail = docList.pop();
            nextTableName = tail.table;
        }
        if (!docList) {
            throw StorageErrorFactory_1.default.getEntityNotFound(context);
        }
        return [docList, nextTableName];
    }
    async insertTableEntity(context, table, account, entity) {
        const tablesCollection = this.db.getCollection(this.TABLES_COLLECTION);
        const tableDocument = tablesCollection.findOne({
            account,
            table
        });
        if (!tableDocument) {
            throw StorageErrorFactory_1.default.getTableNotExist(context);
        }
        const tableEntityCollection = this.db.getCollection(this.getTableCollectionName(account, table));
        if (!tableEntityCollection) {
            throw StorageErrorFactory_1.default.getTableNotExist(context);
        }
        const doc = tableEntityCollection.findOne({
            PartitionKey: entity.PartitionKey,
            RowKey: entity.RowKey
        });
        if (doc) {
            throw StorageErrorFactory_1.default.getEntityAlreadyExist(context);
        }
        entity.properties.Timestamp = utils_1.getTimestampString(entity.lastModifiedTime);
        entity.properties["Timestamp@odata.type"] = "Edm.DateTime";
        tableEntityCollection.insert(entity);
        return entity;
    }
    async insertOrUpdateTableEntity(context, table, account, entity, ifMatch) {
        if (ifMatch === undefined) {
            // Upsert
            const existingEntity = await this.queryTableEntitiesWithPartitionAndRowKey(context, table, account, entity.PartitionKey, entity.RowKey);
            if (existingEntity) {
                // Update
                return this.updateTableEntity(context, table, account, entity, ifMatch);
            }
            else {
                // Insert
                return this.insertTableEntity(context, table, account, entity);
            }
        }
        else {
            // Update
            return this.updateTableEntity(context, table, account, entity, ifMatch);
        }
        // const tablesCollection = this.db.getCollection(this.TABLES_COLLECTION);
        // const tableDocument = tablesCollection.findOne({
        //   account,
        //   table
        // });
        // if (!tableDocument) {
        //   throw StorageErrorFactory.getTableNotExist(context);
        // }
        // const tableEntityCollection = this.db.getCollection(
        //   this.getTableCollectionName(account, table)
        // );
        // if (!tableEntityCollection) {
        //   throw StorageErrorFactory.getTableNotExist(context);
        // }
        // const doc = tableEntityCollection.findOne({
        //   PartitionKey: entity.PartitionKey,
        //   RowKey: entity.RowKey
        // }) as Entity;
        // if (!doc) {
        //   throw StorageErrorFactory.getEntityNotExist(context);
        // } else {
        //   // Test if etag value is valid
        //   if (ifMatch === "*" || doc.eTag === ifMatch) {
        //     tableEntityCollection.remove(doc);
        //     entity.properties.Timestamp = getTimestampString(
        //       entity.lastModifiedTime
        //     );
        //     entity.properties["Timestamp@odata.type"] = "Edm.DateTime";
        //     tableEntityCollection.insert(entity);
        //     return;
        //   }
        // }
        // throw StorageErrorFactory.getPreconditionFailed(context);
    }
    async insertOrMergeTableEntity(context, table, account, entity, ifMatch) {
        if (ifMatch === undefined) {
            // Upsert
            const existingEntity = await this.queryTableEntitiesWithPartitionAndRowKey(context, table, account, entity.PartitionKey, entity.RowKey);
            if (existingEntity) {
                // Merge
                return this.mergeTableEntity(context, table, account, entity, ifMatch);
            }
            else {
                // Insert
                return this.insertTableEntity(context, table, account, entity);
            }
        }
        else {
            // Merge
            return this.mergeTableEntity(context, table, account, entity, ifMatch);
        }
        // const tablesCollection = this.db.getCollection(this.TABLES_COLLECTION);
        // const tableDocument = tablesCollection.findOne({
        //   account,
        //   table
        // });
        // if (!tableDocument) {
        //   throw StorageErrorFactory.getTableNotExist(context);
        // }
        // const tableEntityCollection = this.db.getCollection(
        //   this.getTableCollectionName(account, table)
        // );
        // if (!tableEntityCollection) {
        //   throw StorageErrorFactory.getTableNotExist(context);
        // }
        // const doc = tableEntityCollection.findOne({
        //   PartitionKey: entity.PartitionKey,
        //   RowKey: entity.RowKey
        // }) as Entity;
        // if (!doc) {
        //   throw StorageErrorFactory.getEntityNotExist(context);
        // } else {
        //   // Test if etag value is valid
        //   if (ifMatch === "*" || doc.eTag === ifMatch) {
        //     const mergedDEntity: Entity = {
        //       ...doc,
        //       ...entity,
        //       properties: {
        //         // TODO: Validate incoming odata types
        //         ...doc.properties,
        //         ...entity.properties,
        //         Timestamp: getTimestampString(entity.lastModifiedTime),
        //         "Timestamp@odata.type": "Edm.DateTime"
        //       },
        //       lastModifiedTime: context.startTime!
        //     };
        //     tableEntityCollection.remove(doc);
        //     tableEntityCollection.insert(mergedDEntity);
        //     return mergedDEntity;
        //   }
        // }
        // throw StorageErrorFactory.getPreconditionFailed(context);
    }
    async deleteTableEntity(context, table, account, partitionKey, rowKey, etag) {
        const tablesCollection = this.db.getCollection(this.TABLES_COLLECTION);
        const tableDocument = tablesCollection.findOne({
            account,
            table
        });
        if (!tableDocument) {
            throw StorageErrorFactory_1.default.getTableNotExist(context);
        }
        const tableEntityCollection = this.db.getCollection(this.getTableCollectionName(account, table));
        if (!tableEntityCollection) {
            throw StorageErrorFactory_1.default.getTableNotExist(context);
        }
        if (partitionKey !== undefined && rowKey !== undefined) {
            const doc = tableEntityCollection.findOne({
                PartitionKey: partitionKey,
                RowKey: rowKey
            });
            if (!doc) {
                throw StorageErrorFactory_1.default.getEntityNotFound(context);
            }
            else {
                if (etag !== "*" && doc.eTag !== etag) {
                    throw StorageErrorFactory_1.default.getPreconditionFailed(context);
                }
            }
            tableEntityCollection.remove(doc);
            return;
        }
        throw StorageErrorFactory_1.default.getPropertiesNeedValue(context);
    }
    async queryTableEntities(context, account, table, queryOptions, nextPartitionKey, nextRowKey) {
        const tablesCollection = this.db.getCollection(this.TABLES_COLLECTION);
        const tableDocument = tablesCollection.findOne({
            account,
            table
        });
        if (!tableDocument) {
            throw StorageErrorFactory_1.default.getTableNotExist(context);
        }
        const tableEntityCollection = this.db.getCollection(this.getTableCollectionName(account, table));
        if (!tableEntityCollection) {
            throw StorageErrorFactory_1.default.getTableNotExist(context);
        }
        let queryWhere;
        try {
            queryWhere = this.generateQueryEntityWhereFunction(queryOptions.filter);
        }
        catch (e) {
            throw StorageErrorFactory_1.default.getQueryConditionInvalid(context);
        }
        const segmentFilter = {};
        if (nextPartitionKey) {
            segmentFilter.PartitionKey = { $gte: nextPartitionKey };
        }
        if (nextRowKey) {
            segmentFilter.RowKey = { $gte: nextRowKey };
        }
        const maxResults = queryOptions.top || constants_1.QUERY_RESULT_MAX_NUM;
        const result = tableEntityCollection
            .chain()
            .find(segmentFilter)
            .where(queryWhere)
            .sort((obj1, obj2) => {
            if (obj1.PartitionKey > obj2.PartitionKey) {
                return 1;
            }
            else if (obj1.PartitionKey === obj2.PartitionKey) {
                if (obj1.RowKey > obj2.RowKey) {
                    return 1;
                }
                else if (obj1.RowKey === obj2.RowKey) {
                    return 0;
                }
                else {
                    return -1;
                }
            }
            else {
                return -1;
            }
        })
            .limit(maxResults + 1)
            .data();
        let nextPartitionKeyResponse;
        let nextRowKeyResponse;
        if (result.length > maxResults) {
            const tail = result.pop();
            nextPartitionKeyResponse = tail.PartitionKey;
            nextRowKeyResponse = tail.RowKey;
        }
        return [result, nextPartitionKeyResponse, nextRowKeyResponse];
    }
    async queryTableEntitiesWithPartitionAndRowKey(context, table, account, partitionKey, rowKey) {
        const tableColl = this.db.getCollection(this.getTableCollectionName(account, table));
        // Throw error, if table not exists
        if (!tableColl) {
            throw StorageErrorFactory_1.default.getTableNotExist(context);
        }
        // Get requested Doc
        const requestedDoc = tableColl.findOne({
            PartitionKey: partitionKey,
            RowKey: rowKey
        });
        return requestedDoc;
    }
    async getTableAccessPolicy(context, table) {
        // TODO
        throw new NotImplementedError_1.default(context);
    }
    async setTableAccessPolicy(context, table) {
        // TODO
        throw new NotImplementedError_1.default(context);
    }
    async updateTableEntity(context, table, account, entity, ifMatch) {
        const tablesCollection = this.db.getCollection(this.TABLES_COLLECTION);
        const tableDocument = tablesCollection.findOne({
            account,
            table
        });
        if (!tableDocument) {
            throw StorageErrorFactory_1.default.getTableNotExist(context);
        }
        const tableEntityCollection = this.db.getCollection(this.getTableCollectionName(account, table));
        if (!tableEntityCollection) {
            throw StorageErrorFactory_1.default.getTableNotExist(context);
        }
        const doc = tableEntityCollection.findOne({
            PartitionKey: entity.PartitionKey,
            RowKey: entity.RowKey
        });
        if (!doc) {
            throw StorageErrorFactory_1.default.getEntityNotFound(context);
        }
        // Test if etag value is valid
        const encodedEtag = doc.eTag.replace(":", "%3A").replace(":", "%3A");
        let encodedIfMatch;
        if (ifMatch !== undefined) {
            encodedIfMatch = ifMatch.replace(":", "%3A").replace(":", "%3A");
        }
        if (encodedIfMatch === undefined ||
            encodedIfMatch === "*" ||
            (encodedIfMatch !== undefined && encodedEtag === encodedIfMatch)) {
            tableEntityCollection.remove(doc);
            entity.properties.Timestamp = utils_1.getTimestampString(entity.lastModifiedTime);
            entity.properties["Timestamp@odata.type"] = "Edm.DateTime";
            tableEntityCollection.insert(entity);
            return entity;
        }
        throw StorageErrorFactory_1.default.getPreconditionFailed(context);
    }
    async mergeTableEntity(context, table, account, entity, ifMatch) {
        const tablesCollection = this.db.getCollection(this.TABLES_COLLECTION);
        const tableDocument = tablesCollection.findOne({
            account,
            table
        });
        if (!tableDocument) {
            throw StorageErrorFactory_1.default.getTableNotExist(context);
        }
        const tableEntityCollection = this.db.getCollection(this.getTableCollectionName(account, table));
        if (!tableEntityCollection) {
            throw StorageErrorFactory_1.default.getTableNotExist(context);
        }
        const doc = tableEntityCollection.findOne({
            PartitionKey: entity.PartitionKey,
            RowKey: entity.RowKey
        });
        if (!doc) {
            throw StorageErrorFactory_1.default.getEntityNotFound(context);
        }
        // if match is URL encoded from the clients, match URL encoding
        // this does not always seem to be consisten...
        const encodedEtag = doc.eTag.replace(":", "%3A").replace(":", "%3A");
        let encodedIfMatch;
        if (ifMatch !== undefined) {
            encodedIfMatch = ifMatch.replace(":", "%3A").replace(":", "%3A");
        }
        if (encodedIfMatch === undefined ||
            encodedIfMatch === "*" ||
            (encodedIfMatch !== undefined && encodedEtag === encodedIfMatch)) {
            const mergedDEntity = Object.assign(Object.assign(Object.assign({}, doc), entity), { properties: Object.assign({}, doc.properties
                // ...entity.properties
                ) });
            // Merge inner properties
            for (const key in entity.properties) {
                if (Object.prototype.hasOwnProperty.call(entity.properties, key)) {
                    if (key.endsWith(constants_1.ODATA_TYPE)) {
                        continue;
                    }
                    const value = entity.properties[key];
                    mergedDEntity.properties[key] = value;
                    if (entity.properties[`${key}${constants_1.ODATA_TYPE}`] !== undefined) {
                        mergedDEntity.properties[`${key}${constants_1.ODATA_TYPE}`] =
                            entity.properties[`${key}${constants_1.ODATA_TYPE}`];
                    }
                    else {
                        delete mergedDEntity.properties[`${key}${constants_1.ODATA_TYPE}`];
                    }
                }
            }
            tableEntityCollection.update(mergedDEntity);
            return mergedDEntity;
        }
        else {
            throw StorageErrorFactory_1.default.getPreconditionFailed(context);
        }
    }
    getTableCollectionName(account, table) {
        return `${account}$${table}`;
    }
    static tokenizeQuery(originalQuery) {
        // Escape a single backtick to prevent interpreting the start of a template literal.
        const query = originalQuery.replace(/`/g, "\\`");
        let tokenStart = 0;
        const tokens = [];
        let inString = false;
        let i;
        function appendToken() {
            if (i - tokenStart > 0) {
                let token;
                if (inString) {
                    // Extract the token and unescape quotes
                    token = query.substring(tokenStart, i).replace(/''/g, "'");
                    // Extract the leading type prefix, if any.
                    const stringStart = token.indexOf("'");
                    const typePrefix = token.substring(0, stringStart);
                    const backtickString = "`" + token.substring(typePrefix.length + 1) + "`";
                    // Remove the GUID type prefix since we compare these as strings
                    if (typePrefix === "guid") {
                        token = backtickString;
                    }
                    else {
                        token = typePrefix + backtickString;
                    }
                }
                else {
                    token = convertToken(query.substring(tokenStart, i));
                }
                if (token) {
                    tokens.push(token);
                }
            }
            tokenStart = i + 1;
        }
        function convertToken(token) {
            switch (token) {
                case "TableName":
                    return "name";
                case "eq":
                    return "===";
                case "gt":
                    return ">";
                case "ge":
                    return ">=";
                case "lt":
                    return "<";
                case "le":
                    return "<=";
                case "ne":
                    return "!==";
                case "and":
                    return "&&";
                case "or":
                    return "||";
                case "not":
                    return "!";
                default:
                    return token;
            }
        }
        for (i = 0; i < query.length; i++) {
            if (inString) {
                // Look for a double quote, inside of a string.
                if (i < query.length - 1 && query[i] === "'" && query[i + 1] === "'") {
                    i++;
                    continue;
                }
                else if (query[i] === "'") {
                    appendToken();
                    inString = false;
                }
            }
            else if (query[i] === "(" || query[i] === ")") {
                i--;
                appendToken();
                i++;
                tokens.push(query[i]);
                tokenStart++;
            }
            else if (/\s/.test(query[i])) {
                appendToken();
            }
            else if (query[i] === "'") {
                inString = true;
            }
        }
        appendToken();
        return tokens;
    }
    /**
     * Azurite V2 query entities implementation as temporary workaround before new refactored implementation of querying.
     * TODO: Handle query types
     */
    static transformQuery(query) {
        // If a token is neither a number, nor a boolean, nor a string enclosed with quotation marks it is an operand.
        // Operands are attributes of the object used within the where clause of LokiJS, thus we need to prepend each
        // attribute with an object identifier 'item.attribs'.
        let transformedQuery = "return ( ";
        let isOp = false;
        let previousIsOp = false;
        const tokens = LokiTableMetadataStore.tokenizeQuery(query);
        let counter = -1;
        for (let token of tokens) {
            counter++;
            if (token === "") {
                continue;
            }
            previousIsOp = isOp;
            isOp = ["===", ">", ">=", "<", "<=", "!=="].includes(token);
            if (!token.match(/\b\d+/) &&
                token !== "true" &&
                token !== "false" &&
                !token.includes("`") &&
                ![
                    "===",
                    ">",
                    ">=",
                    "<",
                    "<=",
                    "!==",
                    "&&",
                    "||",
                    "!",
                    "(",
                    ")"
                ].includes(token)) {
                if (token === "PartitionKey" || token === "RowKey") {
                    transformedQuery += `item.${token} `;
                }
                else {
                    // Datetime compare
                    if (counter + 2 <= tokens.length - 1 &&
                        tokens[counter + 2].startsWith("datetime")) {
                        transformedQuery += `new Date(item.properties.${token}).getTime() `;
                    }
                    else {
                        transformedQuery += `item.properties.${token} `;
                    }
                }
            }
            else {
                // Remove "L" from long int
                // 2039283L ==> 2039283
                const matchLongInt = token.match(/\b[0-9]*L\b/g);
                if (previousIsOp &&
                    matchLongInt !== null &&
                    matchLongInt.length === 1) {
                    token = token.replace(/L\b/g, "");
                    // however, as long int is stored as string, we need to add inverted commas
                    token = "'" + token + "'";
                }
                else if (previousIsOp && token.startsWith("datetime")) {
                    token = token.replace(/\bdatetime\b/g, "");
                    token = `new Date(${token}).getTime()`;
                }
                else if (previousIsOp &&
                    (token.startsWith("X") || token.startsWith("binary"))) {
                    throw Error("Binary filter is not supported yet.");
                }
                transformedQuery += `${token} `;
            }
        }
        transformedQuery += ")";
        return transformedQuery;
    }
    /**
     * @param query Query Enties $query string.
     */
    generateQueryEntityWhereFunction(query) {
        if (query === undefined) {
            return () => true;
        }
        const transformedQuery = LokiTableMetadataStore.transformQuery(query);
        // tslint:disable-next-line: no-console
        // console.log(query);
        // tslint:disable-next-line: no-console
        // console.log(transformedQuery);
        return new Function("item", transformedQuery);
    }
}
exports.default = LokiTableMetadataStore;
//# sourceMappingURL=LokiTableMetadataStore.js.map